// GraphPlot.java
// Aug 6, 2002; Oct 30, 2002, June 30, 2003, Mar 23, 2005
//
// This draws the detail screens for the DataView project.
// This manages the details of drawing, loading data from database,
// resizing the picture, etc.
//
// Overview: 
//   show a list of route descriptions on a left screen.
//   The user selects routes by moving them to the right screen.
//   That operation reads the route data from zip file, loads a
//   database table with gps point information and draws the
//   set of points to the screen.
//   The user may choose a color for each data set.
//   The user may delete any individual data set from the drawing,
//   which deletes its rows from the database.
//   The user may zoom in or change the center of the entire display.
//


import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;                 // Color
import java.awt.event.*;           // mouseListener
import java.util.*;                // Vector
import javax.swing.event.*;        // ListSelectionListener
import java.sql.*;
import java.io.*;

//-------------------------------------------------
// New improved plotting ideas?
//
// 3. alt vs. time plotting: 
// 4. want alt vs. distance plotting (sqrt(lat^2 + lon^2))?
// 5. want distance vs. time
// 6. want total altitude climbed
// 8. want slow plot to trace a route as travelled
//
// separate the initial file selector/deselector
// and color selector from the selected data display/zoom utility?
// they seem to be fairly separable entities.
//
//=======================================================

public class GraphPlot  extends JPanel
 implements MouseListener, MouseMotionListener, ListSelectionListener {

  JLabel label1;
  JLabel labelRun;
  JButton qButton;
  JCheckBox checkBox;
  JLabel labelplotTop, labelplotBot, labelplotLeft, labelplotRight;
  JLabel labeldataTop, labeldataBot, labeldataLeft, labeldataRight;
  JLabel mouseLat, mouseLon;
  JLabel loadStats, totalStats;
  final JLabel plotLabel;
  JComboBox trekBox;
  JComboBox colorBox;
  JComboBox loadedTreks;
  JPanel trekPanel;
  JScrollPane availableScroll, loadedScroll, colorScroll;
  JTextField dataField;
  JRadioButton redButton, blueButton, blackButton, cyanButton, whiteButton;
  ButtonGroup group;

// these are used by the repaint() function dispatcher:

  int mode = 0;
  final int CLEAR = 1;
  final int GPS_A = 8;
  final int GPS_B = 9;
  final int GPS_C = 10;

// general plotting stuff

  int plotStart, plotEnd, displayEdgeWidth;
  int plotmaxlat, plotminlat, plotmaxlon, plotminlon;
  int plotmaxdate, plotmindate, plotstatus;
  int maxlat, minlat, maxlon, minlon;
  int minPlotlatindex, maxPlotlatindex, minPlotlonindex, maxPlotlonindex;
  int displayCount;
  int maxHeight, maxWidth;
  boolean rawParsed, rangeset;

// the list of datasets
 
  gpsdata gps;

// various plot data arrays

  int Xscaled[], Yscaled[], Xdisplay[], Ydisplay[];
  int latFiltered[], lonFiltered[], altFiltered[];
  Color latlonColor[];

// the route selection and color collection

  Vector availableData, loadedData, colorData;
  JList availableRoutes, loadedRoutes, colorList;
  int routeIndex;

  Color currentColor;

  Connection conn;
  static PreparedStatement prepAdd;
  static PreparedStatement prepRmv;
  static PreparedStatement prepIndex;
  static PreparedStatement prepColor;

//--------------------------------------------------
// This constructor builds things that probably won't 
// change during the program run.
//---------------------------------------------------

  GraphPlot() { 


    plotStart=0;
    plotEnd = getWidth();

    displayEdgeWidth = 20;

    plotLabel = new JLabel("buffer empty");

    conn = null;

    try{
      Class.forName("org.hsql.jdbcDriver");                 // load JDBC

      conn=DriverManager.getConnection(
        "jdbc:HypersonicSQL:routeDB","sa",""
      );
    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

    gps = new gpsdata(conn);           // load route description and geocache list
   
    try{
      prepAdd=conn.prepareCall("INSERT INTO display (route) VALUES (?) ");
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

    try{
      prepRmv=conn.prepareCall("DELETE FROM display WHERE route = ? ");
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

   try{
      prepRmv=conn.prepareCall("UPDATE RouteData set color = ? where Route = ? ");
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

    plotinit();
    mouseInit();
  }

//----------------------------------------------------
// "plotinit" refreshes things that may need to be called again during a run.
//----------------------------------------------------

  public void plotinit(){

    rawParsed = false;

// initially these are set to extremes so they don't restrict plot limits

    plotmaxlat = plotmaxlon = plotmaxdate = (int) Integer.MAX_VALUE;
    plotminlat = plotminlon = plotmindate = (int) Integer.MIN_VALUE;

// these are initialized to the opposite limits to make sure that any
// possible GPS data will override existing limits and reset the value.

    maxlat = maxlon = (int) Integer.MIN_VALUE;
    minlat = minlon = (int) Integer.MAX_VALUE;

    plotstatus = 1;
    displayCount = 0;

    rangeset = false;
   
    minPlotlatindex = maxPlotlatindex = minPlotlonindex = maxPlotlonindex = 0;

  }

//-----------------------------------------------------

  public JPanel createButtonArea1() {

 
    JPanel panel = new JPanel();
    panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS ));
    panel.setBorder(new BevelBorder( BevelBorder.RAISED));
    panel.setMaximumSize(new Dimension(700,50));
 
// make the route management scrollpanes

    availableScroll = new JScrollPane();
    availableScroll.setPreferredSize(new Dimension(200,100));
 
    loadedScroll = new JScrollPane();
    loadedScroll.setPreferredSize(new Dimension(200,100));

// build both lists and populate the left side

    reloadList();

// make the control panel to sit between the two lists

    JPanel trekControl = new JPanel();
    trekControl.setLayout(new BoxLayout(trekControl, BoxLayout.Y_AXIS ));
 
    int width = 60;  
    int height = 24;  

    JButton button;

    button = new JButton("Add >");
    trekControl.add(button);
    button.setPreferredSize(new Dimension(width,height));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){ 
        datasetAdd();
      }
    });

    button = new JButton("Rmv <");
    trekControl.add(button);
    button.setPreferredSize(new Dimension(width,height));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){ 
        datasetRmv();
      }
    });
 
    button = new JButton("Rmv All");
    trekControl.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){ 
        datasetRmvAll();
      }
    });

// make a panel to hold these lists, load it and add to the main panel.

    trekPanel = new JPanel();
    trekPanel.setLayout(new BoxLayout(trekPanel, BoxLayout.X_AXIS ));

    width = 700; 
    height = 150;
    trekPanel.setPreferredSize(new Dimension(width,height));

    trekPanel.add(availableScroll);
    trekPanel.add(trekControl);
    trekPanel.add(loadedScroll);
  
    panel.add(trekPanel);

    return panel;
  }

//---------------------------------

  public void datasetAdd(){

    int selection = availableRoutes.getSelectedIndex();

    if (selection < 0) return;

// Get index of selected route

    routeIndex = selection;
    String route = (String) availableRoutes.getSelectedValue();

// lookup routeIndex from database using this descr

    try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery(
      "SELECT route FROM Routes where descr = " + "'" + route + "'");
     if(result.next()) {
       routeIndex = result.getInt(1);
       System.out.println("sql returns Index: " + routeIndex);
     }
    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

//  System.out.println("Index: " + routeIndex);

// list display: add Descr to right:
 
    loadedData.addElement(route);
    loadedRoutes.setListData(loadedData);
    loadedScroll.revalidate();
    loadedScroll.repaint();

// list display: remove Descr from left:

    availableData.removeElementAt(selection);
    availableRoutes.setListData(availableData);
    availableScroll.revalidate();
    availableScroll.repaint();

// Add this route to the "display" table

    try{
      prepAdd.clearParameters();
      prepAdd.setInt(1, routeIndex); 
      prepAdd.execute();
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

// plot display:

    graphGPS_parseRaw();                   // load the selected file(s)
    graphGPS_A();                          // display the current array
        
  }

//---------------------------------------------------------------------

  public void datasetRmv(){

    int selection = loadedRoutes.getSelectedIndex();

    if (selection < 0) return;

    String route = (String) loadedRoutes.getSelectedValue();

// Add this descr back to the left side

    availableData.addElement(route);
    availableRoutes.setListData(availableData);
    availableScroll.revalidate();
    availableScroll.repaint();

// and remove it from the right side

    loadedData.removeElementAt(selection);
    loadedRoutes.setListData(loadedData);
    loadedScroll.revalidate();
    loadedScroll.repaint();

// Remove this route from the "display" table


// lookup routeIndex from database using this descr

    try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery(
      "SELECT route FROM Routes where descr = " + "'" + route + "'");
     if(result.next()) {
       routeIndex = result.getInt(1);
       System.out.println("sql returns Index: " + routeIndex);
     }
    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

// then using index, remove from Display

    try{
      prepRmv.clearParameters();
      prepRmv.setInt(1, routeIndex); 
      prepRmv.execute();
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

// debug:

   System.out.println("List remaining Routes in Display");
 
   try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery("SELECT route FROM Display");
     while(result.next()) {
       int rt = result.getInt(1);
       System.out.println("Routes: " + rt);
     }
   } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

  System.out.println("END List remaining Routes in Display");
 
    rangeset = false;
    graphGPS_A();                          // display the current array
        
  }

//----------------------------------------------------

  public void datasetRmvAll(){
    plotinit();
    graphClear();
    reloadList();
  }

//-------------------------------------------------------------------
// This is the main call to load file data into the routeData table.
// First check whether the data is already loaded...

  public boolean graphGPS_parseRaw(){

  int loaded = 0;

  try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery(
      "SELECT count(*) FROM RouteData where route = " + routeIndex );
     if(result.next()) {
       loaded = result.getInt(1);
     }
   } catch(Exception e) { System.out.println(e); e.printStackTrace(); }


//   System.out.println("ParseRaw is Loading index " + routeIndex);


    Color c = getUserColor();
    int clr = c.getRGB();

   if (loaded == 0){
     gps.loadFiles(routeIndex, c);
   } else {

/*
    try{
      prepColor.clearParameters();
      prepColor.setInt(1, clr); 
      prepColor.setInt(2, routeIndex); 
      prepColor.execute();
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }
*/

     System.out.println("already in database");
   }

    rangeset = false;    // force a resetting of the plotting limits
                         // every time a new route is loaded?
    return rawParsed;
  }

//-------------------------------------------------------
// The main display lat-lon master routine.
// This is called after every change to the screen -
//   adding data, resizing or moving display

  public void graphGPS_A(){

    populatePlotArray();     // filter raw data into latFiltered[], lonFiltered[]
    scaleLatLonArray();      // scales those arrays to fit plot
  }

//---------------------------------------------------------
// In populatePlotArray(), copy data from the database table
// into the "filtered" plot array, based on preset limits.
// Initially the limits default to all data. For the
// next passes, the user may adjust the limits either greater or lesser
// than the lat and lon limits of the original data.
//----------------------------------------------------------

  public void populatePlotArray(){

// First traverse the raw data arrays to get the limited counts...

    displayCount = 0;

// After new data has been added, clear the range setting: 
// open up the filtering to include every possible value.
 
    if (rangeset == false){
      plotmaxlat = plotmaxlon = plotmaxdate = (int) Integer.MAX_VALUE;
      plotminlat = plotminlon = plotmindate = (int) Integer.MIN_VALUE;
    }

// get number of points from database

   try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery(
      "SELECT count(*) FROM RouteData, Display where RouteData.route = Display.route " +
      " and   lat > " + plotminlat +
      " and   lat < " + plotmaxlat +
      " and   lon > " + plotminlon +
      " and   lon < " + plotmaxlon );
     if(result.next()) {
       displayCount = result.getInt(1);
     }
   
   } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

//  System.out.println("Resultcount: " + displayCount );

    if (displayCount == 0){ return; }


// Make some new arrays sized by that count...

    latFiltered = new int[displayCount];
    lonFiltered = new int[displayCount];
    latlonColor = new Color[displayCount];
    altFiltered = new int[displayCount]; 

  
// Populate the selected data from the full data arrays
// While we're at it, also find max and min values.

   maxlat = maxlon = (int) Integer.MIN_VALUE;
   minlat = minlon = (int) Integer.MAX_VALUE;

// also save the filtered array index max and mins so we can identify 
// the display limits in the information boxes below the display.

   int lat, lon, alt, color;


   try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery(
      "SELECT lat, lon, alt, color FROM RouteData, Display where RouteData.route = Display.route " +
      " and   lat > " + plotminlat +
      " and   lat < " + plotmaxlat +
      " and   lon > " + plotminlon +
      " and   lon < " + plotmaxlon );

    int j = 0;

    while(result.next()) {

          lat = result.getInt("lat");
          lon = result.getInt("lon");
          alt = result.getInt("alt");
          color = result.getInt("color");

//  System.out.println("Lat: " + lat + " lon: " + lon + " alt: " + alt );

          latFiltered[j] = lat;
          lonFiltered[j] = lon;
          altFiltered[j] = alt;

          if (latFiltered[j] < minlat){ minlat = latFiltered[j]; minPlotlatindex = j; }
          if (latFiltered[j] > maxlat){ maxlat = latFiltered[j]; maxPlotlatindex = j; }
 
          if (lonFiltered[j] < minlon){ minlon = lonFiltered[j]; minPlotlonindex = j; }
          if (lonFiltered[j] > maxlon){ maxlon = lonFiltered[j]; maxPlotlonindex = j; }

          latlonColor[j] = new Color(color);
          j++;
      }

    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

// How do we adjust the min and max values to keep the picture
// so it represents the physical aspect ratio of the actual land area?
// If the data is wider that it is tall, then want to add "dummy" amounts
// to the height (lat), and if it is taller, then want to pad the longitude.
// 1. what is the data aspect ratio?
// 2. what is the screen display aspect ratio?
// 3. what is the physical aspect ratio?

// At 37 degrees of latitude,
// A minute of Latitude  = 1.15 miles, or 6072 feet.
// A minute of Longitude ~ 0.91 miles, or 4805 feet.

// It looks like the display pixel dimensions are physically the same: 
// 200 pixels measures 18 cm both vertically and horizontally.
// Accident? Design? I don't know.

// So determine the scale factor for both Lat and Lon,
// both in one-ten thousandth of a minute of the measure.

    int latRange = maxlat - minlat;
    int lonRange = maxlon - minlon;

// To be correctly scaled, the latRange/lonRange must equal the ratio 6072/4805.

    double latLonRatio = (double)latRange/(double)lonRange;
    double requiredRatio = (double)6072/(double)4805;

// If the existing ratio is greater than ideal then Lon must be enlarged,
// or else lat must be enlarged.

    if (latLonRatio > requiredRatio){
      int enlarge = (int) ( (double) lonRange * latLonRatio/requiredRatio);
      maxlon += enlarge >> 1;
      minlon -= enlarge >> 1;
    } else {
      int enlarge = (int) ( (double) latRange * latLonRatio/requiredRatio);
      maxlat += enlarge >> 1;
      minlat -= enlarge >> 1;
    }

// Zoom in, out, and image shifting modify these following values,
// so only initialize them once.

    if (rangeset == false){
      plotminlat = minlat;
      plotminlon = minlon;
      plotmaxlat = maxlat;
      plotmaxlon = maxlon;
      rangeset = true;

      System.out.println("PopulatePlotArray: rescale...");
      System.out.println("min Lat: " + plotminlat);
      System.out.println("min Lon: " + plotminlon);
      System.out.println("max Lat: " + plotmaxlat);
      System.out.println("max Lon: " + plotmaxlon);
    }


  //  redisplay();   // refresh the informational text display

  }

//---------------------------------------------------------
// By the time scalePlotArray is called, the raw data files 
// have been selected by parseRaw(), the data array has been
// populated by populatePlotArray(), and the plot limits
// may have been adjusted to display other than the data max/min.
// scalePlotArray uses the selected data and scale factors to
// draw to the display.
//------------------------------------------------------------

  public void scaleLatLonArray(){

    int x,y;

// horizontal is Longitude; vertical is Latitude

    int w = getWidth() - displayEdgeWidth;
    double lonscale  = ((double)w)  / ((double)(plotmaxlon - plotminlon));

    int h = getHeight() - displayEdgeWidth;
    double latscale  = ((double)h) / ((double)(plotmaxlat - plotminlat));
 
    int edge = displayEdgeWidth/2;

    Graphics graphicsForDrawing; 
    graphicsForDrawing = getGraphics();

    Color color = getUserColor();
    graphicsForDrawing.setColor(color);

    maxHeight = getHeight();
    maxWidth = getWidth();

// this line does a fast display erase

    graphicsForDrawing.clearRect(edge, edge, getWidth() - edge, getHeight() - edge);

    Xscaled = new int[displayCount];
    Yscaled = new int[displayCount];

    for(int i = 0; i < displayCount; i++){
    
      x = Xscaled[i] = edge + (int) (((double) (lonFiltered[i] - plotminlon)) * lonscale);
      y = Yscaled[i] = edge + (int) (((double) (latFiltered[i] - plotminlat)) * latscale); 

      if (latlonColor[i] != color){
        color = latlonColor[i];
        graphicsForDrawing.setColor(color);
      }

      if (checkBox.isSelected() == false){
        if ((x > 0) && (x < maxWidth) && (y > 0) && (y < maxWidth))
          graphicsForDrawing.drawLine(x, y, x, y); 
      }

    }

// when fastDrawing, request that the Xscaled[], Yscaled arrays be drawn by the OS.

    if (checkBox.isSelected() == true){
      mode = GPS_A;
      repaint();
    }

    graphicsForDrawing.dispose();

}

//-----------------------------------------------------------------
// copy of similar function in class gpsdata

 public static int dms2n(String position){

    if (position.length() < 8) return 0;

    int ptindex = position.indexOf(".");
    String dd = position.substring(0, ptindex-2);
    String mmffff = position.substring(ptindex-2);
    double frac = (60 * Double.parseDouble(dd)) + Double.parseDouble(mmffff);

    return (int)(10000 * frac);
  }

//------------------------------------------------------------
// Place an X on the screen with the given Lat and Lon coordinates.
// desc is not used in the pgm; but is useful as documentation.
 
  public void scalePlotX(int lat, int lon, String desc){

    double latscale, lonscale;
    int x,y, edge, crosslen;

    crosslen = 5;  

    lonscale  = ((double)getWidth())  / ((double) (plotmaxlon - plotminlon)); 
    latscale  = ((double)getHeight()) / ((double) (plotmaxlat - plotminlat));
  
    Graphics graphicsForDrawing; 
    graphicsForDrawing = getGraphics();

    Color color = Color.black;
    graphicsForDrawing.setColor(color);

    edge = displayEdgeWidth/2;

    x = (int) (((double) (lon-plotminlon)) * lonscale);
    y = (int) (((double) (lat-plotminlat)) * latscale); 

    if ((x > edge) && (x+crosslen < getWidth()-edge) && 
        (y > edge) && (y+crosslen < getHeight()-edge))
    {
      graphicsForDrawing.drawLine(x-crosslen, y, x+crosslen, y); 
      graphicsForDrawing.drawLine(x, y-crosslen, x, y+crosslen); 
    }
   
    graphicsForDrawing.dispose();

//    try{ Thread.sleep(1); } catch(InterruptedException ex){}


  }



//-----------------------------------------------------
// reloadList() is used both to initially build the route list
// and to refresh the lists during program usage.

  void reloadList(){

    availableData = new Vector();
    
  try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery("SELECT Descr FROM Routes");
    while(result.next()) { availableData.addElement(result.getString(1)); }
  } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

    availableRoutes = new JList(availableData);
    availableRoutes.addListSelectionListener(this);
    availableRoutes.setListData(availableData);
    availableScroll.getViewport().add(availableRoutes);

    availableScroll.revalidate();
    availableScroll.repaint();

    loadedData = new Vector();

    loadedRoutes = new JList(loadedData);
    loadedRoutes.setListData(loadedData);
    loadedScroll.getViewport().add(loadedRoutes);

    loadedScroll.revalidate();
    loadedScroll.repaint();
  }

//-----------------------------------------------------

  public void valueChanged(ListSelectionEvent event){

    if (event.getSource() == availableRoutes && 
        !event.getValueIsAdjusting() &&
         availableRoutes.getSelectedValue() != null){

      String route = (String) availableRoutes.getSelectedValue();
      System.out.println("Selected " + route);

    }
  }

//-----------------------------------------------------


  public JPanel createColorPanel() {

    int width = 60;  // width of "<", ">", color radiobutton controls
    int height = 24;  // 26 too big

// the color... Want to show the user the color of each plot...

    JPanel panel = new JPanel();
    panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS ));

    redButton = new JRadioButton("red", true);

    redButton.setPreferredSize(new Dimension(width,height));
    blueButton = new JRadioButton("blue", false);
    blueButton.setPreferredSize(new Dimension(width,height));
    blackButton = new JRadioButton("black", false);
    blackButton.setPreferredSize(new Dimension(width,height));
    cyanButton = new JRadioButton("cyan", false);
    cyanButton.setPreferredSize(new Dimension(width,height));

    checkBox = new JCheckBox("fast draw");
    checkBox.setSelected(true);
    panel.add(checkBox);
    panel.add(redButton);
    panel.add(blueButton);
    panel.add(blackButton);
    panel.add(cyanButton);

    group = new ButtonGroup();
    group.add(redButton);
    group.add(blueButton);
    group.add(blackButton);
    group.add(cyanButton);
    return panel;
  }

  Color getUserColor() {

    if (redButton.isSelected()) return Color.red;
    if (blueButton.isSelected()) return Color.blue;
    if (blackButton.isSelected()) return Color.black;
    if (cyanButton.isSelected()) return Color.cyan;
    return Color.white;
  }


//###########################################################################

  public JPanel createButtonArea2() {

    JPanel panel = new JPanel();
    panel.setLayout(new BorderLayout());
    panel.setBorder(new BevelBorder( BevelBorder.RAISED));
 
    panel.add(leftPanel(), BorderLayout.WEST);
    panel.add(middlePanel(), BorderLayout.CENTER);
    panel.add(rightPanel(), BorderLayout.EAST);

    return panel;
  }

//------------------------------------------------

  public JPanel leftPanel() {

    JPanel panel = new JPanel();
    panel.setLayout(new GridLayout(3,1));           // 3 vertical objects
    panel.setBorder(new BevelBorder( BevelBorder.RAISED));
 
    JButton button;
 
    button = new JButton("Long vs. Lat");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        graphGPS_A();
      }
    });

    button = new JButton("Alt vs. time");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        graphGPS_B();
      }
    });
 
    button = new JButton("Alt vs. Distance");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        graphGPS_C();
      }
    });

    return panel;
  }

//------------------------------------------------

  public JPanel middlePanel() {

    JPanel panel = new JPanel();
    panel.setLayout(new BorderLayout());
   
    JButton button;
  
    button = new JButton("Shift up");
    panel.add(button, BorderLayout.NORTH);
    button.setPreferredSize(new Dimension(100,40));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        shiftUp();
      }
    });


    button = new JButton("Shift left");
    panel.add(button, BorderLayout.WEST);
    button.setPreferredSize(new Dimension(100,30));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        shiftLeft();
      }
    });

    panel.add(zoomPanel(), BorderLayout.CENTER);

    button = new JButton("Shift right");
    panel.add(button, BorderLayout.EAST);
    button.setPreferredSize(new Dimension(100,30));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        shiftRight();
      }
    });
 
    button = new JButton("Shift down");
    panel.add(button, BorderLayout.SOUTH);
    button.setPreferredSize(new Dimension(100, 40));
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        shiftDown();
      }
    }); 

    return panel;
  }

//------------------------------------------------

  public JPanel zoomPanel() {

    JPanel panel = new JPanel();
    panel.setLayout(new GridLayout(2,1));               // 2 vertical objects

    panel.setBorder(new CompoundBorder(new EtchedBorder(), 
      new LineBorder(Color.blue)) );

    JButton button;

    button = new JButton("Zoom in");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        zoomIn();
      }
    });

    button = new JButton("Zoom out");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        zoomOut();
      }
    });

    return panel;
  }

//------------------------------------------------

  public JPanel rightPanel() {

    JPanel panel = new JPanel();
    panel.setLayout(new GridLayout(3,1));             // 3 vertical objects
    panel.setBorder(new BevelBorder( BevelBorder.RAISED));
 
    JButton button;

    button = new JButton("Clear");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        graphClear();
      }
    });


    return panel;
  }

//###########################################################################

// Shift amount: was 10.0 for button shifting; lets try 20.0 for mouse shifting

  public void shiftLeft(){
    if (rangeset){

          double lonrange = plotmaxlon - plotminlon;
          double lonrangetenth = lonrange/20.0;
          plotmaxlon = plotmaxlon + (int) lonrangetenth;
          plotminlon = plotminlon + (int) lonrangetenth;

          graphGPS_A();
        }
  }

  public void shiftRight(){
    if (rangeset){

          double lonrange = plotmaxlon - plotminlon;
          double lonrangetenth = lonrange/20.0;
          plotmaxlon = plotmaxlon - (int) lonrangetenth;
          plotminlon = plotminlon - (int) lonrangetenth;

          graphGPS_A();
        }
  }

  public void shiftUp(){
    if (rangeset){

          double latrange = plotmaxlat - plotminlat;
          double latrangetenth = latrange/20.0;
          plotmaxlat = plotmaxlat + (int) latrangetenth;
          plotminlat = plotminlat + (int) latrangetenth;

          graphGPS_A();
        }
  }

  public void shiftDown(){
    if (rangeset){

          double latrange = plotmaxlat - plotminlat;
          double latrangetenth = latrange/20.0;
          plotmaxlat = plotmaxlat - (int) latrangetenth;
          plotminlat = plotminlat - (int) latrangetenth;

          graphGPS_A();
        }
  }

  public void zoomIn(){
   if (rangeset){

          int latrange = plotmaxlat - plotminlat;
          int sixteenth = latrange >> 4;
          plotmaxlat = plotmaxlat - sixteenth;
          plotminlat = plotminlat + sixteenth;

          int lonrange = plotmaxlon - plotminlon;
          sixteenth = lonrange >> 4;
          plotmaxlon = plotmaxlon - sixteenth;
          plotminlon = plotminlon + sixteenth;

          graphGPS_A();

        }
  }

  public void zoomOut(){
    if (rangeset){

          int latrange = plotmaxlat - plotminlat;
          int sixteenth = latrange >> 4;
          plotmaxlat = plotmaxlat + sixteenth;
          plotminlat = plotminlat - sixteenth;

          int lonrange = plotmaxlon - plotminlon;
          sixteenth = lonrange >> 4;
          plotmaxlon = plotmaxlon + sixteenth;
          plotminlon = plotminlon - sixteenth;

          graphGPS_A();
        }
  }

//###########################################################################
   
//-----------------------------------------------------
// called by tabscreens(). 

  public JPanel createButtonArea5() {
 
    JButton button;

    JPanel panel = new JPanel();

    panel.setLayout(new GridLayout(1,5));             // 5 horizontal objects

    panel.setBorder(new BevelBorder( BevelBorder.RAISED));

//    labelplotTop = new JLabel("watch me"); 
//    panel.add(labelplotTop);

//    labelplotLeft = new JLabel(); 
//    panel.add(labelplotLeft);

//    labelplotBot = new JLabel(); 
//    panel.add(labelplotBot);
 
//    labelplotRight = new JLabel();  
//    panel.add(labelplotRight);

    mouseLat = new JLabel("Latitude");
//    mouseLat.setBorder(new BevelBorder( BevelBorder.RAISED));
    panel.add(mouseLat);

    mouseLon = new JLabel();
    panel.add(mouseLon);

    totalStats = new JLabel();
    panel.add(totalStats);

    loadStats = new JLabel();
    panel.add(loadStats);


    button = new JButton("halt");
    panel.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e){
        System.exit(0);
      }
    });

//    panel.setMaximumSize(new Dimension(700,50));

    return panel;
  }

//-------------------------------------------

  String to4DecPlcStr(int tenthousanths){

    String s1, deg, mins, s2, s3, s4, s5;
    int degrees, ind;
    double d1, minutes;

// because we have ranged and offset the plotting values, we need to save
// the indexes of the min and max, then go back to the original data arrays.
// The original arrays store the data in ten-thousanths of a minute, so we
// need to multiply this back up to minutes, then get the degree part out.

    d1 = (double) tenthousanths;  
    d1 = d1 / 10000;                 
    degrees = (int) (d1/60);
    minutes = d1 - (degrees * 60);
    deg = new Integer(degrees).toString();

    s5 = new Double(minutes).toString();
    ind = 5 + s5.indexOf('.',0);
    if (s5.length() >= ind)
      mins = s5.substring(0, ind);
    else
      mins = s5;

    return deg + " " + mins;
  }

//-------------------------------------------------------
// Alt vs. time 

 public void graphGPS_B(){

    populatePlotArray();   // filter raw data into latFiltered[], lonFiltered[]

    scaleAltArray();      // scales those arrays to fit plot
  }

//--------------------------------------------------------

public void scaleAltArray(){

    double altscale;
    double temp;
    int h,w;

// Altitude vs. time is a different animal than Lon vs. Lat. We need
// to condense maybe 9000 altitude readings into maybe 400 horiz pts.
// Strictly speaking, time is another set of data points, but
// for all practical porpoises, the array index may be taken as 
// equivilent to a linear sequence of time.
// Also, before the GPS receiver acquires sync, it emits NULL altitudes
// (but it emits saved latitudes and longitudes).
// So lets discard those NULLs.

// 1. elim null altitudes and subtract offsets

    Xscaled = new int[displayCount];
    int alts = 0;
 
    for(int i = 0; i < displayCount; i++){

      if (altFiltered[i] == -1) continue;

      Xscaled[alts] = altFiltered[i]; // - gps.minalt;
      alts++;
    }

// Horizontal scaling: example: 9360 points, 586 pixels. 16 points per pixel.

    w = getWidth() - displayEdgeWidth;
    int hscale = alts/w;

    Yscaled = new int[alts];

    int i,j,horz,sums;
    int fullstepper = 0;
    int shortstepper = 0;

// Ahh, don't want fullstepper to access past alts...

    while(fullstepper < alts){
      sums = 0;

      for(j=0; j<hscale; j++)
        if (fullstepper < alts)
          sums = sums + Xscaled[fullstepper++];

      Yscaled[shortstepper++] = sums;
    }
  
// Ok, so we have a Yscaled array of 0 to getWidth(), about.
// Now scale the values to fit the display.

   int tmpmin, tmpmax;

    tmpmin = tmpmax = Yscaled[0];

    for(i = 0; i < shortstepper; i++){
      if (Yscaled[i] < tmpmin) tmpmin = Yscaled[i];      
      if (Yscaled[i] > tmpmax) tmpmax = Yscaled[i];      
    }
 
    for(i = 0; i < shortstepper; i++){
      Yscaled[i] = Yscaled[i] - tmpmin;      
    }
 
// scale to fit the current display

    double scale;

    h = getHeight() - displayEdgeWidth;
    scale  = ((double)h) / ((double)(tmpmax - tmpmin));
 
    for(i = 0; i < shortstepper; i++){

      temp = (double) Yscaled[i];
      Yscaled[i] = (int) (temp * scale);
    }
     
    mode = GPS_B;

    repaint();
  }

//-------------------------------------------------------
// Alt vs. distance

 public void graphGPS_C(){

    populatePlotArray();   // filter raw data into latFiltered[], lonFiltered[]

    scaleAltDistArray();      // scales those arrays to fit plot
  }

//-------------------------------------------------------

public void scaleAltDistArray(){

    double altscale;
    double temp;
    int h,w;

// Altitude vs. time is a different animal than Lon vs. Lat. We need
// to condense maybe 9000 altitude readings into maybe 400 horiz pts.
// Strictly speaking, time is another set of data points, but
// for all practical porpoises, the array index may be taken as 
// equivilent to a linear sequence of time.
// Also, before the GPS receiver acquires sync, it emits NULL altitudes
// (but it emits saved latitudes and longitudes).
// So lets discard those NULLs.

// 1. elim null altitudes and subtract offsets

    Xscaled = new int[displayCount];
    int alts = 0;
 
    for(int i = 0; i < displayCount; i++){

      if (altFiltered[i] == -1) continue;

      Xscaled[alts] = altFiltered[i]; // - gps.minalt;
      alts++;
    }

// Horizontal scaling: example: 9360 points, 586 pixels. 16 points per pixel.

    w = getWidth() - displayEdgeWidth;
    int hscale = alts/w;

    Yscaled = new int[alts];

    int i,j,horz,sums;
    int fullstepper = 0;
    int shortstepper = 0;

// Ahh, don't want fullstepper to access past alts...

    while(fullstepper < alts){
      sums = 0;

      for(j=0; j<hscale; j++)
        if (fullstepper < alts)
          sums = sums + Xscaled[fullstepper++];

      Yscaled[shortstepper++] = sums;
    }
  
// Ok, so we have a Yscaled array of 0 to getWidth(), about.
// Now scale the values to fit the display.

   int tmpmin, tmpmax;

    tmpmin = tmpmax = Yscaled[0];

    for(i = 0; i < shortstepper; i++){
      if (Yscaled[i] < tmpmin) tmpmin = Yscaled[i];      
      if (Yscaled[i] > tmpmax) tmpmax = Yscaled[i];      
    }
 
    for(i = 0; i < shortstepper; i++){
      Yscaled[i] = Yscaled[i] - tmpmin;      
    }
 
// scale to fit the current display

    double scale;

    h = getHeight() - displayEdgeWidth;
    scale  = ((double)h) / ((double)(tmpmax - tmpmin));
 
    for(i = 0; i < shortstepper; i++){

      temp = (double) Yscaled[i];
      Yscaled[i] = (int) (temp * scale);
    }
     
    mode = GPS_C;

    repaint();
  }

//-------------------------------------------------------------------
// clear screen inside borders. 

 public void graphClear(){

    mode = CLEAR;
    repaint();
  }

//---------------------------------------------------------
// This is called by the VM:

  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    currentScreen(g);
  }

// and the VM in turn calls one of:
//-------------------------------------------------------------

  private void currentScreen(Graphics g){

    switch(mode){

      case CLEAR:
            clearScreen(g);
            break;
      case GPS_A:
            plotGPS_A(g);
            break;
      case GPS_B:
            plotGPS_B(g);
            break;
      case GPS_C:
            plotGPS_C(g);
            break;

        default:
            return;
      }
  }
 
//-----------------------
// clear screen

    private void clearScreen(Graphics g){
      int edge;
      edge = displayEdgeWidth/2;

      g.clearRect(edge, edge, getWidth() - edge, getHeight() - edge);
    }

//-----------------------
// Lat vs. Lon

  private void plotGPS_A(Graphics g){

    int i, x, y, edge;
    Color color;

    edge = displayEdgeWidth/2;
    color = Color.blue;
    g.setColor(color);

    g.clearRect(edge, edge, getWidth() - edge, getHeight() - edge);

    for(i = 0; i < Xscaled.length; i++){
 
      x = Xscaled[i];
      y = Yscaled[i];

      if ((x < edge) || (x > maxWidth) || (y < edge) && (y > maxWidth)) continue;

      if (latlonColor[i] != color){
        color =  latlonColor[i];
        g.setColor(color);
      }

      g.drawLine(x,y,x,y);

//      try{ Thread.sleep(1); } catch(InterruptedException ex){}

    }
  }

//-----------------------
// Alt vs. time

  private void plotGPS_B(Graphics g){

    int maxWidth = getWidth();
    int maxHeight = getHeight();
    int bigX = getWidth();
    int bigY = getHeight();
    int thisX, nextX, thisY, nextY, edge;

    edge = displayEdgeWidth/2;

    g.setColor(Color.blue);

    thisY = maxHeight - (edge + Yscaled[0]);
    thisX = 0;

    if (thisY < 0 ) thisY = edge;
    else if ( thisY > bigY) thisY = bigY - edge;

    for(nextX = 1; nextX< maxWidth; nextX++){

      nextY = maxHeight - (edge + Yscaled[nextX]);
     
      if (nextY < 0 ) nextY = edge;
      else if ( nextY > bigY) nextY = bigY - edge;

      g.drawLine(thisX,thisY,nextX,nextY);

      thisX = nextX;
      thisY = nextY;

    }
  }

//-----------------------
// Alt vs. distance

  private void plotGPS_C(Graphics g){

    int maxWidth = getWidth();
    int maxHeight = getHeight();

    int thisX, nextX, thisY, nextY, edge;

    edge = displayEdgeWidth/2;

    g.setColor(Color.blue);

    thisY = maxHeight - (edge + Yscaled[0]);
    thisX = 0;

    for(nextX = 1; nextX< maxWidth; nextX++){

      nextY = maxHeight - (edge + Yscaled[nextX]);
     
      g.drawLine(thisX,thisY,nextX,nextY);

      thisX = nextX;
      thisY = nextY;

    }
  }

//---------------------------------------------------------
// implement the mouse interface

  String msg = "";
  int mouseX = 0, mouseY = 0, button, oldX=0, oldY=0;

  public void mouseInit() {
     addMouseListener(this);
     addMouseMotionListener(this);
  }

  public void mouseClicked(MouseEvent me) {
    mouseX = 0;
    mouseY = 10;
    button = me.getButton();
    msg = "Mouse clicked:  " ;
    System.out.println(msg + button);
  }

  public void mouseEntered(MouseEvent me) {
    mouseX = 0;
    mouseY = 10;
    msg = "Mouse entered.";
    System.out.println(msg);
  }
  public void mouseExited(MouseEvent me) {
    mouseX = 0;
    mouseY = 10;
    msg = "Mouse exited.";
    System.out.println(msg);
  }
  public void mousePressed(MouseEvent me) {
    mouseX = me.getX();
    mouseY = me.getY();
    button = me.getButton();
    msg = "Mouse Pressed:  " ;

// left (1): zoom in; right (3) zoom out

    if (button == 1) zoomIn();
    if (button == 3) zoomOut();

//    System.out.println(msg + button);
  }

  public void mouseReleased(MouseEvent me) {
    mouseX = me.getX();
    mouseY = me.getY();
    button = me.getButton();
    msg = "Mouse released:  " ;
    System.out.println(msg + button);
  }

  public void mouseDragged(MouseEvent me) {
    mouseX = me.getX();
    mouseY = me.getY();

    int diffX = mouseX - oldX;
    int diffY = mouseY - oldY;
 
// lets find the greatest motion - N, E, S or W, and shift that direction.

    diffY = diffY * diffY;
    diffX = diffX * diffX;

    if (diffX > diffY)
      if (mouseX > oldX) shiftRight(); else shiftLeft();
    else
     if (mouseY > oldY) shiftDown(); else shiftUp();
 
    oldX = mouseX;
    oldY = mouseY;

//    msg = "Mouse dragged " ;
//    System.out.println(msg);
  }

//---------------------------------------------------

  public void mouseMoved(MouseEvent me) {

    int x = me.getX(); 
    int y = me.getY();

    if (x < 0 || x > maxWidth || y < 0 || y > maxHeight) return;

    double factor;

// interpolate between current max and mins

    int rangeLat = latFiltered[maxPlotlatindex] - latFiltered[minPlotlatindex];
    factor = (double) rangeLat * (double) y / (double) maxHeight;
    int mLat =  latFiltered[minPlotlatindex] + (int) factor;
    mouseLat.setText("Ptr Lat: " + to4DecPlcStr(mLat));

    int rangeLon = lonFiltered[maxPlotlonindex] - lonFiltered[minPlotlonindex];
    factor = (double) rangeLon * (double) x / (double) maxWidth;
    int mLon =  lonFiltered[minPlotlonindex] + (int) factor;
    mouseLon.setText("Ptr Lon: " + to4DecPlcStr(mLon));
  }

//---------------------------------------------------

}   // end of class

