// gpsdata.java
// July 2, 2003
//
// This class will read a directory full of GPS data files 
// and populate tables with LAT, LON, TIME, DATE, velocity, altitude data.
//
// 1. initial call searches a directory for all data files having 
//    a first line containing "Descr", and will load that filename,
//    description, and index into a "Routes" table.
//
// 2. As user selects individual descriptions, the loader here
//    will be called. It will load all data for a single route
//    into the "routeData" table (if not yet loaded.)
//
// 3. The display class will load the route index being displayed
//    into the "Display" table, and will remove the index if that
//    selection is removed form display by the user.


import java.io.*;
import java.util.*;
import java.sql.*;
import java.awt.*;                 // Color

//-------------------------------------------------------

public class gpsdata {

  public

   static int maxlat, minlat, maxlon, minlon, maxalt, minalt;
   static int routesIndex;
   static int cIndex, cdate, ctime, clon, clat, calt, cspeed;
   static String cfileName, cdescr;
   static String datapath;
   static Connection conn;
   static PreparedStatement routeDataInsert;
   static PreparedStatement routesInsert;
   static Color color;

//--------------------------------------------------

  gpsdata(Connection cn){

    conn = cn;                                   // database connection 
 
//    datapath = "C:\\loggerData\\javaCooked\\";
    datapath = "tripdata\\";

    Statement stat = null;
    String statement;


    try{
      stat=conn.createStatement();
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    try{
      stat.executeUpdate("DROP TABLE RouteData");
    } catch(SQLException e) { }                   // ignore exception - OK if no table.
    

    statement = "CREATE TABLE RouteData (route int, date int, time int, lat int, ";
    statement = statement +         " lon int, alt int, speed int, color int)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    statement = "CREATE INDEX RouteIndex on RouteData (route)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

    statement = "CREATE INDEX latIndex on RouteData (lat)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

    statement = "CREATE INDEX lonIndex on RouteData (lon)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }

    statement = "CREATE INDEX altIndex on RouteData (alt)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    try{
      stat.executeUpdate("DROP TABLE Display");
    } catch(SQLException e) { }                   // ignore exception - OK if no table.


    statement = "CREATE TABLE Display (route int)";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    try{
      stat.executeUpdate("DROP TABLE Routes");
    } catch(SQLException e) { }                   // ignore exception - OK if no table.


    statement = "CREATE TABLE Routes(route int,filename varchar(80),descr varchar(80))";
    try{
      stat.execute(statement);
      stat.close(); 
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    statement = "INSERT INTO RouteData(route,date,time,lat,lon,alt,speed,color)";
    statement = statement + "VALUES(?,?,?,?,?,?,?,?)";
    try{
      routeDataInsert=conn.prepareCall(statement);
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    statement = "INSERT INTO Routes(route,filename,descr)VALUES(?,?,?)";
    try{
      routesInsert=conn.prepareCall(statement);
    } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }


    routesIndex = 0;

    loadRouteList();       
 }

//-------------------------------------------------
// database insertion routines. used down below, but here because when modifying 
// the table format, the "create table", the "prepare SQL" above and the "setInt..." 
// below must agree exactly.
 

  static void insertRoutes() {
  
    try{
      routesInsert.clearParameters();
      routesInsert.setInt(1, routesIndex); 
      routesInsert.setString(2, cfileName); 
      routesInsert.setString(3, cdescr); 
      routesInsert.execute();
     } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }
  }


  static void insertRouteData() {
  
    try{
      routeDataInsert.clearParameters();
      routeDataInsert.setInt(1, cIndex); 
      routeDataInsert.setInt(2, cdate); 
      routeDataInsert.setInt(3, ctime); 
      routeDataInsert.setInt(4, clat); 
      routeDataInsert.setInt(5, clon); 
      routeDataInsert.setInt(6, calt); 
      routeDataInsert.setInt(7, cspeed); 
      routeDataInsert.setInt(8, color.getRGB()); 
      routeDataInsert.execute();
     } catch(SQLException e) {   System.out.println(e); e.printStackTrace(); }
  }


//----------------------------------------------------
// This is called once by the class constructor.
// Load "Routes" table with all located route file descriptions, names and generate indexes.

  static void loadRouteList(){

    String listName;
    String thisLine;

    routesIndex = 0;                     // This will be the route index.

    File file = new File(datapath);  
    String thisFile;
    File[] files = file.listFiles();
    int filenameLength;

// List all files in directory. Read first line from each file.

    for (int i = 0; i < files.length; i++) {
 
      thisFile = files[i].getName(); 

      try {

        BufferedReader br = new BufferedReader(new FileReader(datapath + thisFile)); 
        if ((thisLine = br.readLine()) != null){
          if (thisLine.length() < 10) continue;
          if (thisLine.substring(0,5).compareTo("Descr") != 0) continue;

          cfileName = datapath + thisFile;
          cdescr = thisLine.substring(5);

          insertRoutes();

          routesIndex++;
        }
      } catch (IOException e) { System.err.println("Exception: " + e); }
    }

/*
// debug: show what we've got

   int i;
   String s;
   try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery("SELECT route, descr FROM Routes");
    while(result.next()) {
       i = result.getInt(1);
       s = result.getString(2);
       System.out.println("Index " + i + " descr: <" + s + ">");
     }
    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

*/


  }

//-----------------------------------------------
// This is called every time the user asks for a route that hasn't yet
// been loaded. It looks up the filaname based on the index then loads the 
// "RouteData" table from that data file.

public boolean loadFiles(int index, Color c){

  color = c;
  String s = "";

  try{
    Statement stat=conn.createStatement();
    ResultSet result=stat.executeQuery("SELECT fileName FROM Routes where route = " + index);
    if(result.next()) { s = result.getString(1); }
  } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

  gpsDatProcess(s, index);
  updateStatistics();
  return true;

}

//--------------------------------------------------
// New format: file may contain a few leading lines of desc.
// Data is fixed length.

void gpsDatProcess(String fileName, int index){

    String thisLine, field;  
    int ptr, size;

    try {
      BufferedReader br = new BufferedReader(new FileReader(fileName));
 
      while ((thisLine = br.readLine()) != null) { 

        if (thisLine.substring(0,5).compareTo("Descr") == 0) continue;

        size = thisLine.length();
        if (size != 36) {
          System.out.println("Odd size line: " + size);
          continue;
        }

//050806 14782 26145977 31345280 1047 0126

        cIndex = index;

        ptr = 0; size = 6;
        field = thisLine.substring( ptr,  ptr+size);
//      System.out.println("<" + field + ">");   System.exit(0);
        cdate  = myInt(field); 
        ptr += size;

        size = 5;
        field = thisLine.substring( ptr,  ptr+size);
//      System.out.println("<" + field + ">");  System.exit(0);
        ctime  = myInt(field); 
        ptr += size;

        size = 8;
        field = thisLine.substring( ptr,  ptr+size);
//      System.out.println("Lat <" + field + ">");  System.exit(0);
        clat  = myInt(field); 
        ptr += size;

        size = 8;
        field = thisLine.substring( ptr,  ptr+size);
//        System.out.println("Lon <" + field + ">");  // System.exit(0);
        clon  = myInt(field); 
        ptr += size;

        size = 4;
        field = thisLine.substring( ptr,  ptr+size);
//      System.out.println("<" + field + ">");  System.exit(0);
        calt  = myInt(field); 
        ptr += size;

        size = 4;
        field = thisLine.substring( ptr,  ptr+size);
//      System.out.println("<" + field + ">");  System.exit(0);
        cspeed  = myInt(field); 
 
        insertRouteData();

//        System.out.println(ctime + " " + clat + " " + clon);
      }


     } catch (IOException e) { System.err.println("Exception: " + e); }

}

//-------------------------------------------------------------------

  public int myInt(String field){

    int myNumber;

    try{
      myNumber  = Integer.parseInt(field);
    } catch(java.lang.NumberFormatException e)
    {
        System.out.println("Int conversion error: <" + field + ">");
        myNumber = -1;
    }

    return myNumber;
  }
 
//---------------------------------------------------
// updateStatistics() finds the max and min lat and lon in the arrays.
// These are needed for plotting because while time is monotonic, 
// the path travelled over time is not.

  void updateStatistics() {

  minlon = maxlon = minlat = maxlat = 5;

  Statement stat = null;
  ResultSet result = null;
  String query;

  try{
    stat=conn.createStatement();

//    stat.execute("DELETE FROM Routes where lon < 73220000"); 
//    stat.execute("DELETE FROM Routes where lon > 73230000"); 
//    stat.execute("DELETE FROM Routes where lat < 22750000"); 
//    stat.execute("DELETE FROM Routes where lat > 22760000"); 

    stat=conn.createStatement();
    query = "SELECT min(lat) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);  
    if(result.next()) { minlat = result.getInt(1); }

    stat=conn.createStatement();
    query = "SELECT max(lat) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);
    if(result.next()) { maxlat = result.getInt(1); }

    stat=conn.createStatement();
    query = "SELECT min(lon) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);
    if(result.next()) { minlon = result.getInt(1); }

    stat=conn.createStatement();
    query = "SELECT max(lon) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);
    if(result.next()) { maxlon = result.getInt(1); }

    stat=conn.createStatement();
    query = "SELECT min(alt) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);
    if(result.next()) { minalt = result.getInt(1); }

    stat=conn.createStatement();
    query = "SELECT max(alt) FROM RouteData, Display where RouteData.route = Display.route";
    result=stat.executeQuery(query);
    if(result.next()) { maxalt = result.getInt(1); }

    } catch(Exception e) { System.out.println(e); e.printStackTrace(); }

  status();

  }

// ----------------------------
 
  public void status(){
 
    System.out.println("gpsdata status:");
    System.out.println("routes         " + routesIndex);
    System.out.println("min lon        " + minlon);
    System.out.println("max lon        " + maxlon);
    System.out.println("min lat        " + minlat);
    System.out.println("max lat        " + maxlat);
    System.out.println("min alt        " + minalt);
    System.out.println("max alt        " + maxalt);
  }

} 


//--the end -------------------------------------
